/**************************************************
 *
  File: mc-agent.cc
  Author: Suman Banerjee <suman@cs.umd.edu>
  Date: 15th March, 2001
  Terms: GPL

  Narada implementation in myns

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
**************************************************/

#include <stdlib.h>
#include <assert.h>
#include <scheduler.h>
#include <node.h>
#include "mc-agent.h"
#include "misc.h"

NamedConstant const_seek_nbr_cnt("const_seek_nbr_cnt",8);
NamedConstant const_mesh_neighbors("const_mesh_neighbors",2);
NamedConstant const_mesh_neighbor_keepalive_period("const_mesh_neighbor_keepalive_period",30.00);
NamedConstant const_mesh_neighbor_refresh_timeout("const_mesh_neighbor_refresh_timeout",60.00);
NamedConstant const_mesh_rt_exchange_period("const_mesh_rt_exchange_period",30.00);
NamedConstant const_mesh_rt_exchange_timeout("const_mesh_rt_exchange_timeout",60.00);
NamedConstant const_mesh_link_add_period("const_mesh_link_add_period",100.00);
NamedConstant const_mesh_link_drop_period("const_mesh_link_drop_period",100.00);
NamedConstant const_mesh_setup_neighbor_timeout("const_mesh_setup_neighbor_timeout",10.00);
NamedConstant const_data_transmit_interval("const_data_transmit_interval",2.00);
NamedConstant linkdropthreshold("linkdropthreshold",2);
NamedConstant linkaddthreshold("linkaddthreshold",5.00);
NamedConstant infinitelinkcost("infinitelinkcost",10000.00);
NamedConstant invalid("invalid",-12345);
NamedConstant maxmeshmembers("maxmeshmembers",50);
NamedConstant random_data_source("random_data_source",2);


mcAgent::~mcAgent (void) {

  delete mesh_refresh_timer;
  delete mesh_keepalive_timer;
  //sri's code 
  delete mesh_rt_exchange_timer; 
  delete mesh_rt_exchange_check_timer; 
  delete mesh_link_add_timer; 
  delete mesh_link_drop_timer; 
  delete data_timer; 

/* Suman's correction */
  // printf ("[timer-debug]  at %8.4f < ag %d nd %d > about to delete all timers \n", Scheduler::Clock(), id,n->id);

  for (void *pos = mesh_setup_neighbor_timer_list.GetHeadPosition(); 
             pos != NULL; 
/* Suman's correction */
             /* pos = */ mesh_setup_neighbor_timer_list.GetNext(pos)) {
    MeshSetupNeighborTimer *msnt = mesh_setup_neighbor_timer_list.GetAt(pos); 
    msnt->CancelTimer();
    delete msnt;
  }
  mesh_setup_neighbor_timer_list.RemoveAll(); 
  for (int i = 0; i < MAXMESHMEMBERS; i++) {
    if ((meshrt.valid[i]) && (meshrt.pathlen[i] == 1))
      assert (meshrt.path[i] == NULL);
    if ((meshrt.valid[i]) && (meshrt.pathlen[i] > 1)) {
      free(meshrt.path[i]); 
      meshrt.path[i] = NULL;
    }
  }

  delete[] meshrt.valid;
  delete[] meshrt.cost;
  delete[] meshrt.pathlen;
  delete[] meshrt.path;
  return; 
  //
}

void mcAgent::init (int Id, int Index, Node *N) {

  Agent::init(Id,Index,N);
  t = AGENT_APPLICATION_MC;
  rp.agent_id = -1;
  mesh_refresh_timer = new MeshRefreshTimeoutTimer(this);
  mesh_keepalive_timer = new MeshKeepAliveTimer(this);
//sri's code
  mesh_rt_exchange_timer = new MeshRTExchangeTimer(this);
  mesh_link_add_timer = new MeshLinkAddTimer(this);
  mesh_link_drop_timer = new MeshLinkDropTimer(this);
  mesh_rt_exchange_check_timer = new MeshRTExchangeCheckTimer(this); 
  data_timer = new DataTimer(this); 
  datasource = false;
  // if (id == RANDOM_DATA_SOURCE)  DATA SOURCE not needed -- Suman
  //  datasource = true; 
//
  return;
}

void mcAgent::start (void) {

  #ifdef LOG_MCAGENT_START
  printf("[start] at %8.4f < ag %d nd %d > started\n",Scheduler::Clock(),id,n->id);
  #endif // LOG_MCAGENT_START

  Agent::start();
  meshrt.valid = new (bool[MAXMESHMEMBERS]);
  meshrt.cost = new (double[MAXMESHMEMBERS]);
  meshrt.pathlen = new (int[MAXMESHMEMBERS]);
  meshrt.path= new (AgentInfo * [MAXMESHMEMBERS]);
  //sri's code
  for (int i = 0; i < MAXMESHMEMBERS; i++) {
    meshrt.valid[i] = false; 
    meshrt.cost[i] = (double) INVALID;
    meshrt.pathlen[i] = INVALID; 
    meshrt.path[i] = NULL;
  }
  meshrt.valid[id] = true; 
  meshrt.cost[id] = 0;
  meshrt.pathlen[id] = 0; 
  meshrt.path[id] = NULL;
  //
  if (rp.agent_id == -1) {
    printf ("[Err] at %8.4f < ag %d nd %d > doesn't know RP\n",Scheduler::Clock(),id,n->id);
    exit (-1);
  }

  //sri's code
  data_seq_no = 0; 
  if (datasource == true) 
    data_timer->SetTimer(get_time_till_next_transmit());
  //
  mesh_refresh_seq_no = 0;
  AppPacket *p = new AppPacket(PING_RP);
  p->u.ping_p.seek_nbr_cnt = CONST_SEEK_NBR_CNT;
  send_pkt(p,rp.agent_id,rp.node_id);
  return;
}

void mcAgent::stop (void) {

  #ifdef LOG_MCAGENT_STOP
    printf("[mcagent-stop] at %8.4f < ag %d nd %d > stopped \n",Scheduler::Clock(),id,n->id);
  #endif // LOG_MCAGENT_STOP

  mesh_refresh_timer->CancelTimer();
  mesh_keepalive_timer->CancelTimer();
  //sri's code  
  mesh_rt_exchange_timer->CancelTimer();
  mesh_link_add_timer->CancelTimer();
  mesh_link_drop_timer->CancelTimer();
  mesh_rt_exchange_check_timer->CancelTimer(); 
  data_timer->CancelTimer(); 

/* Suman's correction */
  // printf ("[timer-debug]  at %8.4f < ag %d nd %d > about to delete all timers \n", Scheduler::Clock(), id,n->id);

  for (void *pos = mesh_setup_neighbor_timer_list.GetHeadPosition(); 
             pos != NULL; 
/* Suman's correction */
             /* pos =  mesh_setup_neighbor_timer_list.GetNext(pos) */) {
    MeshSetupNeighborTimer *msnt = mesh_setup_neighbor_timer_list.GetAt(pos); 
    msnt->CancelTimer();
    delete msnt; 
/* Suman's correction -- BEGIN */
    void * old_pos = pos;
    mesh_setup_neighbor_timer_list.GetNext(pos);
    mesh_setup_neighbor_timer_list.RemoveAt(old_pos);
/* END */
  }

  setup_neighbor_list.RemoveAll();
  mesh_neighbor_list.RemoveAll();
  stopped_member_list.RemoveAll();
  mesh_setup_neighbor_timer_list.RemoveAll(); 
  remove_all_agent_list();
  Agent::stop();
  return;
}

void mcAgent::remove_all_agent_list (void) {

  for (void *pos = all_agent_list.GetHeadPosition();
       pos != NULL;
       ) {
    MeshMember *mm = all_agent_list.GetAt(pos);
    delete[] mm->source;
    void *old_pos = pos;
    all_agent_list.GetNext(pos);
    delete mm;
    all_agent_list.RemoveAt(old_pos);
  }
  return;
}

void mcAgent::specific_rx_pkt_handler  (Packet *p) {

  
  if (p->t != PACKET_APP) {

     printf ("[Err] at %8.4f < ag %d nd %d > received unknown packet type\n", Scheduler::Clock(),id,n->id);
     return;
  }

  AppPacket *ap = (AppPacket *)p;
  switch (ap->st) {

  case PING_RP_RESPONSE :

    handle_rp_ping_response(ap);
    break;

  case MESH_NEIGHBOR_SETUP_REQUEST :

    handle_mesh_neighbor_setup_req(ap);
    break;

  case MESH_NEIGHBOR_SETUP_RESPONSE :

    handle_mesh_neighbor_setup_resp(ap);
    break;

  case MESH_NEIGHBOR_REFRESH :
    
    handle_mesh_member_refresh(ap);
    break;

//sri's code
  case MESH_RT_EXCHANGE :

    handle_mesh_rt_exchange(ap);
    break;

  case MESH_LINK_ADD :

    handle_mesh_link_add(ap);
    break; 

  case MESH_LINK_ADD_RESPONSE :
  
    handle_mesh_link_add_response(ap);
    break;

  case MESH_LINK_DROP :
  
    handle_mesh_link_drop(ap);
    break; 

  case DATA :
  
    handle_data(ap); 
    break; 
//

  default :
    printf ("[Err] at %8.4f < ag %d nd %d > received unknown packet subtype\n", Scheduler::Clock(),id,n->id);
  }

  return;
}

MeshMember *mcAgent::create_new_mesh_member_info (int aid, int nid) {

  assert (all_agent_list.Locate(aid) == NULL);
  MeshMember *mm = new MeshMember (aid, nid, Scheduler::Clock());
  all_agent_list.Add(mm,aid);
  return mm;
}

void mcAgent::handle_rp_ping_response (AppPacket *ap) {

  for (int i = 0; i < ap->u.pingresp_p.ret_nbr_cnt; i++) {

    MeshMember *mm = create_new_mesh_member_info(ap->u.pingresp_p.ret_nbrs[i].agent_id, ap->u.pingresp_p.ret_nbrs[i].node_id);

#ifdef LOG_APP_RP
    printf ("[pingresp] at %8.4f < ag %d nd %d > found < ag %d nd %d >\n", Scheduler::Clock(), id, n->id, mm->ainfo.agent_id, mm->ainfo.node_id);
#endif // LOG_APP_RP
  }

  select_mesh_neighbors();
  mesh_refresh_timer->SetTimer(CONST_MESH_NEIGHBOR_REFRESH_TIMEOUT);
  if (mesh_rt_exchange_timer == NULL) {
    printf("[timer-debug] at %8.4f < ag %d nd %d > timer is null\n",Scheduler::Clock(),id,n->id);
  }
  mesh_rt_exchange_timer->SetTimer(CONST_MESH_RT_EXCHANGE_PERIOD); 
  mesh_rt_exchange_check_timer->SetTimer(CONST_MESH_RT_EXCHANGE_TIMEOUT); 
  mesh_link_add_timer->SetTimer(CONST_MESH_LINK_ADD_PERIOD);
  mesh_link_drop_timer->SetTimer(CONST_MESH_LINK_DROP_PERIOD);
  return;
}

/* Select a list of mesh neighbors. These neigbors will be sent */
/* periodic mesh refresh messages */
void mcAgent::select_mesh_neighbors (void) {

  int all_agent_count = all_agent_list.GetSize();
  int num_mesh_neighbors = (CONST_MESH_NEIGHBORS > all_agent_count) ? all_agent_count : CONST_MESH_NEIGHBORS;
  int *neigh_indices = get_rand_set(num_mesh_neighbors,all_agent_count);

  int nlist_index = 0;
  int this_index = 0;
  for (void *pos = all_agent_list.GetHeadPosition();
       pos != NULL;
       all_agent_list.GetNext(pos) ) {

    if (nlist_index == num_mesh_neighbors)
      break;
    MeshMember *mm = all_agent_list.GetAt(pos);
    if (neigh_indices[nlist_index] == this_index) {

    /* Assign this agent to be a neighbor */
      add_to_setup_neighbor_list(mm);
      MeshSetupNeighborTimer *msnt = new MeshSetupNeighborTimer(this,mm);
      msnt->SetTimer(CONST_MESH_SETUP_NEIGHBOR_TIMEOUT); 

/* Suman's correction */
      // printf ("[timer-debug] About to add timer at %8.4f for %d at %d\n", Scheduler::Clock(), mm->ainfo.agent_id, id);

      #ifdef LOG_SELECT_NEIGHBOR 
        printf("[select-neigh] at %8.4f < ag %d nd %d > selects < ag %d nd %d >\n",Scheduler::Clock(),id,n->id,mm->ainfo.agent_id,mm->ainfo.node_id);
      #endif // LOG_SELECT_NEIGHBOR

      mesh_setup_neighbor_timer_list.Add(msnt, mm->ainfo.agent_id); 
      tx_mesh_neighbor_setup_req(mm);
      nlist_index ++;
    }
    this_index ++;
  }
  delete [] neigh_indices;

  // tx_mesh_neighbor_refresh();
  mesh_keepalive_timer->SetTimer(CONST_MESH_NEIGHBOR_KEEPALIVE_PERIOD);
  return;
}

void mcAgent::tx_mesh_neighbor_setup_req (MeshMember *mm) {

  #ifdef LOG_APP_NEIGHBOR
    printf ("[setup-neighbor] at %8.4f < ag %d nd %d > requests < ag %d nd %d >\n", Scheduler::Clock(), id, n->id, mm->ainfo.agent_id, mm->ainfo.node_id);
  #endif // LOG_APP_NEIGHBOR

  AppPacket *ap = new AppPacket (MESH_NEIGHBOR_SETUP_REQUEST);
  send_pkt(ap,mm->ainfo.agent_id,mm->ainfo.node_id);
  return;
}

// Respond to neighbor setup request with accept or reject
void mcAgent::tx_mesh_neighbor_setup_resp (MeshMember *mm, bool accept) {

  AppPacket *ap  = new AppPacket (MESH_NEIGHBOR_SETUP_RESPONSE);
  ap->u.meshneighborsetupresp_p.accept = accept;
  send_pkt(ap,mm->ainfo.agent_id,mm->ainfo.node_id);
  return;
}

void mcAgent::tx_mesh_refresh (void) {

  forward_mesh_refresh(id,n->id,mesh_refresh_seq_no,-1,true);
  return;
}


/* Forward a mesh refresh packet to all neighbors, except the one specified */
void mcAgent::forward_mesh_refresh (int o_aid, int o_nid, int o_seqno, int except_aid, bool isalive) {

  for (void *pos = mesh_neighbor_list.GetHeadPosition();
       pos != NULL;
       mesh_neighbor_list.GetNext(pos) ) {

    MeshMember *mm = mesh_neighbor_list.GetAt(pos);
    if (mm->ainfo.agent_id == except_aid)
      continue;
    AppPacket *ap = new AppPacket(MESH_NEIGHBOR_REFRESH);
    ap->u.meshrefresh_p.is_alive = isalive;
    ap->u.meshrefresh_p.neighbor_seq_no = mesh_refresh_seq_no;
    ap->u.meshrefresh_p.origin.agent_id = o_aid;
    ap->u.meshrefresh_p.origin.node_id = o_nid;
    ap->u.meshrefresh_p.origin_seq_no = o_seqno;

    #ifdef LOG_MESH_REFRESH
      printf ("[mesh-refresh] at %8.4f < ag %d nd %d > is-src < ag %d nd %d > is-dest < ag %d nd %d > is-origin < %d > is-ssn\n", Scheduler::Clock(), id, n->id, mm->ainfo.agent_id, mm->ainfo.node_id, o_aid, o_nid, o_seqno); 
    #endif // LOG_MESH_REFRESH
    
    send_pkt(ap,mm->ainfo.agent_id,mm->ainfo.node_id);
  }
  mesh_refresh_seq_no ++;
  return;
}

//sri's code
void mcAgent::tx_mesh_rt_exchange (void) {
  
  for (void *pos = mesh_neighbor_list.GetHeadPosition();
       pos != NULL;
       mesh_neighbor_list.GetNext(pos) ) {
  
    MeshMember *mm = mesh_neighbor_list.GetAt(pos);
    AppPacket *ap = new AppPacket(MESH_RT_EXCHANGE);
    ap->u.meshrt_p.valid = new (bool[MAXMESHMEMBERS]);
    ap->u.meshrt_p.pathlen = new (int[MAXMESHMEMBERS]);
    ap->u.meshrt_p.cost = new (double[MAXMESHMEMBERS]);
    ap->u.meshrt_p.path = new (struct AgentInfo *[MAXMESHMEMBERS]);
    for (int i = 0; i < MAXMESHMEMBERS; i++) {
      if ((meshrt.valid[i]) && (meshrt.pathlen[i] > 1) && (meshrt.path[i][0].agent_id == mm->ainfo.agent_id)) {
         ap->u.meshrt_p.valid[i] = false; 
         ap->u.meshrt_p.pathlen[i] = INVALID;
         ap->u.meshrt_p.path[i] = NULL;
         ap->u.meshrt_p.cost[i] = INFINITELINKCOST;
      } 
      else {
        ap->u.meshrt_p.valid[i] = meshrt.valid[i];
        ap->u.meshrt_p.pathlen[i] = meshrt.pathlen[i];
        ap->u.meshrt_p.cost[i] = meshrt.cost[i];
        if (meshrt.valid[i] == true) {
          if (i != id)
            if (meshrt.pathlen[i] > 1)
              ap->u.meshrt_p.path[i] = (struct AgentInfo *) safe_malloc((meshrt.pathlen[i]-1)*sizeof(struct AgentInfo));
          for (int j = 0; j < meshrt.pathlen[i] - 1; j++) {
            ap->u.meshrt_p.path[i][j].agent_id = meshrt.path[i][j].agent_id;
            ap->u.meshrt_p.path[i][j].node_id = meshrt.path[i][j].node_id; 
          }
        }
        else {
          ap->u.meshrt_p.path[i] = NULL; 
        }
      }
    }

    #ifdef LOG_APP_MESH_RT 
      printf ("[mesh-rt-send] at %8.4f < ag %d nd %d > updates < ag %d nd %d >\n", Scheduler::Clock(), id, n->id, mm->ainfo.agent_id, mm->ainfo.node_id);
    #endif // LOG_APP_MESH_RT

    send_pkt(ap,mm->ainfo.agent_id,mm->ainfo.node_id);
  }
  return; 
}  

void mcAgent::handle_mesh_rt_exchange(AppPacket *ap) {


  MeshMember *mm = find_mesh_member_info (ap->src_agent);
  bool changes = false;


  // Do not accept anything from non-neighbors
  if (mm == NULL) {

    #ifdef LOG_APP_MESH_RT
      printf ("[mesh-rt-recv] at %8.4f < ag %d nd %d > got-updates-from < ag %d nd %d > begin\n", Scheduler::Clock(), id, n->id, mm->ainfo.agent_id, mm->ainfo.node_id);
    #endif // LOG_APP_MESH_RT

    /*
    for (int i = 0; i < MAXMESHMEMBERS; i++) 
      if ((ap->u.meshrt_p.valid[i] == true) && (ap->u.meshrt_p.pathlen[i] > 1))
        free(ap->u.meshrt_p.path[i]); 
    delete[] ap->u.meshrt_p.path; 
    delete[] ap->u.meshrt_p.valid;
    delete[] ap->u.meshrt_p.pathlen;
    delete[] ap->u.meshrt_p.cost;
    */ 
    return;
  }
  else if (mm->is_neighbor == false) {

    #ifdef LOG_APP_MESH_RT
      printf ("[mesh-rt-recv] at %8.4f < ag %d nd %d > got-updates-from < ag %d nd %d > begin\n", Scheduler::Clock(), id, n->id, mm->ainfo.agent_id, mm->ainfo.node_id);
    #endif // LOG_APP_MESH_RT

    /*
    for (int i = 0; i < MAXMESHMEMBERS; i++) 
      if ((ap->u.meshrt_p.valid[i] == true) && (ap->u.meshrt_p.pathlen[i] > 1))
        free(ap->u.meshrt_p.path[i]); 
    delete[] ap->u.meshrt_p.path; 
    delete[] ap->u.meshrt_p.valid;
    delete[] ap->u.meshrt_p.pathlen;
    delete[] ap->u.meshrt_p.cost; 
    */
    return; 
  }

  #ifdef LOG_APP_MESH_RT
  {
    printf ("[mesh-rt-recv] at %8.4f < ag %d nd %d > got-updates-from < ag %d nd %d > begin\n", Scheduler::Clock(), id, n->id, mm->ainfo.agent_id, mm->ainfo.node_id);
    for (int i = 0; i < MAXMESHMEMBERS; i++) {
      if (ap->u.meshrt_p.valid[i] == true) {
        printf("[route-pkt] to < ag %d > of len %d and cost %8.2f\n",i,ap->u.meshrt_p.pathlen[i],ap->u.meshrt_p.cost[i]);
      }
    }
  } 
#endif // LOG_APP_MESH_RT

  double neighborcost = meshrt.cost[ap->src_agent];
  bool throughme;

  for (int i = 0; i < MAXMESHMEMBERS; i++) {
    if ((meshrt.valid[i]) && (meshrt.pathlen[i] > 1) && (meshrt.path[i][0].agent_id == ap->src_agent)) {
      free(meshrt.path[i]);
      meshrt.path[i] = NULL;
      meshrt.valid[i] = false; 
      if (ap->u.meshrt_p.valid[i]) {
        throughme = false; 
        for (int j = 0; j < ap->u.meshrt_p.pathlen[i] - 1; j++) {
          if (ap->u.meshrt_p.path[i][j].agent_id == id) {
            throughme = true; 
            break; 
          }
        }
        if (!throughme) {
          meshrt.valid[i] = true; 
          meshrt.cost[i] = ap->u.meshrt_p.cost[i] + neighborcost; 
          meshrt.pathlen[i] = ap->u.meshrt_p.pathlen[i] + 1; 
          meshrt.path[i] = (struct AgentInfo *)safe_malloc((meshrt.pathlen[i]-1)*sizeof(struct AgentInfo));
          meshrt.path[i][0].agent_id = ap->src_agent;
          meshrt.path[i][0].node_id = ap->src; 
          for (int j = 1; j < meshrt.pathlen[i] - 1; j++) {
            meshrt.path[i][j].agent_id = ap->u.meshrt_p.path[i][j-1].agent_id; 
            meshrt.path[i][j].node_id = ap->u.meshrt_p.path[i][j-1].node_id; 
          }
        }
      }
    }
    else {
      if (ap->u.meshrt_p.valid[i]) {
        if ((meshrt.valid[i] == false) || (meshrt.cost[i] > ap->u.meshrt_p.cost[i] + neighborcost)) {
          throughme = false; 
          for (int j = 0; j < ap->u.meshrt_p.pathlen[i] - 1; j++) {
            if (ap->u.meshrt_p.path[i][j].agent_id == id) {
              throughme = true; 
              break; 
            }
          }
          if (!throughme) {
            if ((meshrt.valid[i]) && (meshrt.pathlen[i] > 1)) {
              free(meshrt.path[i]);
              meshrt.path[i] = NULL; 
            }
            meshrt.valid[i] = true; 
            meshrt.cost[i] = ap->u.meshrt_p.cost[i] + neighborcost; 
            meshrt.pathlen[i] = ap->u.meshrt_p.pathlen[i] + 1; 
            assert(meshrt.path[i] == NULL);
            meshrt.path[i] = (struct AgentInfo *)safe_malloc((meshrt.pathlen[i]-1)*sizeof(struct AgentInfo));
            meshrt.path[i][0].agent_id = ap->src_agent;
            meshrt.path[i][0].node_id = ap->src; 
            for (int j = 1; j < meshrt.pathlen[i] - 1; j++) {
              meshrt.path[i][j].agent_id = ap->u.meshrt_p.path[i][j-1].agent_id; 
              meshrt.path[i][j].node_id = ap->u.meshrt_p.path[i][j-1].node_id; 
            }
          }
        }
      }
    }
  }
    	  
  mm->gotrt = true; 
  int thrume = 0, thruneighbor = 0; 
  for (int i = 0; i < MAXMESHMEMBERS; i++) {
    MeshMember *src_mm = find_mesh_member_info(i);
    if (meshrt.valid[i])   
      if (meshrt.pathlen[i] > 1)   
        if (meshrt.path[i][0].agent_id == mm->ainfo.agent_id) 
          thruneighbor++;
    if ((ap->u.meshrt_p.pathlen[i] == INVALID) && (ap->u.meshrt_p.valid[i] == false) && (meshrt.valid[i])) {
      thrume++;
      if (src_mm != NULL) {  
        src_mm->source[ap->src_agent] = true; 
      }
    }
    else {
      if (src_mm != NULL)
        src_mm->source[ap->src_agent] = false; 
    }
  }
  if (thrume > thruneighbor)   
    mm->consensuscost = thrume; 
  else 
    mm->consensuscost = thruneighbor; 
  /*
  for (int i = 0; i < MAXMESHMEMBERS; i++) 
    if ((ap->u.meshrt_p.valid[i] == true) && (ap->u.meshrt_p.pathlen[i] > 1))
      free(ap->u.meshrt_p.path[i]); 
  delete[] ap->u.meshrt_p.path; 
  delete[] ap->u.meshrt_p.valid;
  delete[] ap->u.meshrt_p.pathlen;
  delete[] ap->u.meshrt_p.cost; 
  */

  #ifdef LOG_APP_MESH_RT
  {
    printf ("[rt] at %8.4f < ag %d nd %d > after-update-from < ag %d nd %d > begin \n",Scheduler::Clock(),id,n->id,ap->src_agent,ap->src); 
    for (int i = 0; i < MAXMESHMEMBERS; i++) {
      if (meshrt.valid[i] == true) {
        printf("[routing-table] to %d of len %d and cost %8.2f via %d\n",i,meshrt.pathlen[i],meshrt.cost[i],((meshrt.pathlen[i] > 1) ? meshrt.path[i][0].agent_id : -1));
      }
    } 
    printf ("[consensus-cost] to %d is %d\n",mm->ainfo.agent_id,mm->consensuscost);
    printf ("[mesh-rt-recv] at %8.4f < ag %d nd %d > got-updates-from < ag %d nd %d > end\n", Scheduler::Clock(), id, n->id, mm->ainfo.agent_id, mm->ainfo.node_id);
  }
  #endif // LOG_APP_MESH_RT

  return; 

}

void mcAgent::tx_mesh_link_add(void) {
  
  if (all_agent_list.GetSize() == mesh_neighbor_list.GetSize()) {

    #ifdef LOG_LINK_ADD
      printf("[link-add] at %8.4f < ag %d nd %d > has no non-neighbors \n",Scheduler::Clock(),id,n->id);
    #endif // LOG_LINK_ADD

    return;
  }

  int rand = get_rand(all_agent_list.GetSize() - mesh_neighbor_list.GetSize()); 
  void *pos = all_agent_list.GetHeadPosition(); 
  int i = 0; 
  MeshMember *mm = all_agent_list.GetAt(pos);
  while (mm->is_neighbor == true) {
    all_agent_list.GetNext(pos);
    mm = all_agent_list.GetAt(pos);
  }
  while (i < rand) {
    i++;
    all_agent_list.GetNext(pos);
    mm = all_agent_list.GetAt(pos);
    while (mm->is_neighbor == true) {
      all_agent_list.GetNext(pos);
      mm = all_agent_list.GetAt(pos);
    }
  }

  if ((mm->is_neighbor == false) && (mm->neighbor_setup_pending == false)) {
    AppPacket *ap = new AppPacket(MESH_LINK_ADD);

    #ifdef LOG_LINK_ADD
      printf("[link-addreq] at %8.4f < ag %d nd %d > to < ag %d nd %d >\n",Scheduler::Clock(),id,n->id,mm->ainfo.agent_id,mm->ainfo.node_id);
    #endif // LOG_LINK_ADD

    send_pkt(ap,mm->ainfo.agent_id,mm->ainfo.node_id);
  }
  return; 
}

void mcAgent::handle_mesh_link_add (AppPacket *ap) {

  AppPacket *app = new AppPacket(MESH_LINK_ADD_RESPONSE);
  app->u.meshrt_p.valid = new (bool[MAXMESHMEMBERS]);
  app->u.meshrt_p.pathlen = new (int[MAXMESHMEMBERS]);
  app->u.meshrt_p.cost = new (double[MAXMESHMEMBERS]);
  app->u.meshrt_p.path = new (struct AgentInfo *[MAXMESHMEMBERS]);
  for (int i = 0; i < MAXMESHMEMBERS; i++) {
    if ((i != id) && (meshrt.valid[i])) {
      app->u.meshrt_p.valid[i] = meshrt.valid[i];
      app->u.meshrt_p.pathlen[i] = meshrt.pathlen[i];
      app->u.meshrt_p.cost[i] = meshrt.cost[i];
      if (meshrt.pathlen[i] > 1)
        app->u.meshrt_p.path[i] = (struct AgentInfo *)safe_malloc((meshrt.pathlen[i]-1)*sizeof(struct AgentInfo));
      for (int j = 0; j < meshrt.pathlen[i] - 1; j++) {
        app->u.meshrt_p.path[i][j].agent_id = meshrt.path[i][j].agent_id; 
        app->u.meshrt_p.path[i][j].node_id = meshrt.path[i][j].node_id; 
      }
    }
    else {
      app->u.meshrt_p.valid[i] = false;
      app->u.meshrt_p.pathlen[i] = INVALID;
      app->u.meshrt_p.cost[i] = INFINITELINKCOST;
      app->u.meshrt_p.path[i] = NULL; 
    }
  }

  #ifdef LOG_LINK_ADD 
    printf("[link-addresp] at %8.4f < ag %d nd %d > to < ag %d nd %d >\n",Scheduler::Clock(),id,n->id,ap->src_agent,ap->src);
  #endif // LOG_LINK_ADD  

  send_pkt(app,ap->src_agent,ap->src);
  return; 
}  
  
void mcAgent::handle_mesh_link_add_response(AppPacket *ap) {

  #ifdef LOG_LINK_ADD 
    printf("[link-addresp] at %8.4f < ag %d nd %d > from < ag %d nd %d >\n",Scheduler::Clock(),id,n->id,ap->src_agent,ap->src);
  #endif // LOG_LINK_ADD

  double utility=0.0,current,next;
  double neighborcost = n->rt->get_path_cost(ap->src); 
  for (int i = 0; i < MAXMESHMEMBERS; i++) {
    if (i == id) continue; 
    if ((ap->u.meshrt_p.valid[i]) && (meshrt.valid[i])) {
      current = meshrt.cost[i]; 
      next = ap->u.meshrt_p.cost[i] + neighborcost;
      if (next < current) {
	utility += (current - next) / current;

        #ifdef LOG_LINK_ADD
          printf("[link-addresp-details] for %d current cost is %8.2f and next is %8.2f\n",i,current,next);
        #endif // LOG_LINK_ADD

      }
    }
    if ((ap->u.meshrt_p.valid[i]) && (meshrt.valid[i] == false)) { 
      utility += 1.0; 

      #ifdef LOG_LINK_ADD
        printf("[link-addresp-details] for %d new route discovered\n",i);
      #endif // LOG_LINK_ADD

    }
  }

  #ifdef LOG_LINK_ADD
    printf("[link-addresp-details] util is %8.2f\n",utility);
  #endif // LOG_LINK_ADD

  if (utility > LINKADDTHRESHOLD) {

    MeshMember *mm = find_mesh_member_info(ap->src_agent);
    if (mm != NULL) {

      assert (mm->neighbor_setup_pending == false);
      if (mm->is_neighbor == true) {

      #ifdef LOG_LINK_ADD
        printf ("[link-add] at %8.4f < ag %d nd %d > not-needed-to < ag %d nd %d > \n", Scheduler::Clock(),id,n->id, mm->ainfo.agent_id,mm->ainfo.node_id);
      #endif // LOG_LINK_ADD

      }
      else {

        #ifdef LOG_LINK_ADD
          printf("[link-add] at %8.4f < ag %d nd %d > to < ag %d nd %d >\n",Scheduler::Clock(),id,n->id,mm->ainfo.agent_id,mm->ainfo.node_id);
        #endif // LOG_LINK_ADD

        add_to_setup_neighbor_list(mm);
        MeshSetupNeighborTimer *msnt = new MeshSetupNeighborTimer(this,mm);
        msnt->SetTimer(CONST_MESH_SETUP_NEIGHBOR_TIMEOUT);

/* Suman's correction */
      // printf ("[timer-debug] About to add timer at %8.3f for %d at %d\n", Scheduler::Clock(), mm->ainfo.agent_id, id);

        mesh_setup_neighbor_timer_list.Add(msnt, mm->ainfo.agent_id);  
        tx_mesh_neighbor_setup_req(mm);
      }
    }
  }
  /*
  for (int i = 0; i < MAXMESHMEMBERS; i++) 
    if ((ap->u.meshrt_p.valid[i] == true) && (ap->u.meshrt_p.pathlen[i] > 1))
      free(ap->u.meshrt_p.path[i]); 
  delete[] ap->u.meshrt_p.path; 
  delete[] ap->u.meshrt_p.valid;
  delete[] ap->u.meshrt_p.pathlen;
  delete[] ap->u.meshrt_p.cost; 
  */  
  return; 
}

void mcAgent::tx_mesh_link_drop () {

  MeshMember *badneighbor = NULL;
  int badconsensuscost = MAXMESHMEMBERS; 
  for (void *pos = mesh_neighbor_list.GetHeadPosition();
       pos != NULL;
       mesh_neighbor_list.GetNext(pos) ) {

    MeshMember *mm = mesh_neighbor_list.GetAt(pos);
    if (mm->consensuscost < badconsensuscost) {  
      badconsensuscost = mm->consensuscost; 
      badneighbor = mm; 
    }
  }
  
  #ifdef LOG_LINK_DROP 
    if (badneighbor != NULL)
      printf("[link-drop-details] at %8.4f < ag %d nd %d > thinks < ag %d nd %d > has bad consensus cost of %d\n",Scheduler::Clock(),id,n->id,badneighbor->ainfo.agent_id,badneighbor->ainfo.node_id,badconsensuscost);
  #endif // LOG_LINK_DROP
  
  if (badconsensuscost < LINKDROPTHRESHOLD) {
    AppPacket *ap = new AppPacket(MESH_LINK_DROP);
    
    #ifdef LOG_LINK_DROP
      printf("[link-drop] at %8.4f < ag %d nd %d > to < ag %d nd %d >\n",Scheduler::Clock(),id,n->id,badneighbor->ainfo.agent_id,badneighbor->ainfo.node_id);
    #endif // LOG_LINK_DROP
    
    send_pkt(ap,badneighbor->ainfo.agent_id, badneighbor->ainfo.node_id);
    purge_from_neighbor_list(badneighbor);
    cleanup_rt(badneighbor->ainfo.agent_id);
  }
  return; 
}    
  
void mcAgent::handle_mesh_link_drop(AppPacket *ap) {
  
  MeshMember *mm = find_mesh_member_info(ap->src_agent);
  if ((mm != NULL) && (mm->is_neighbor)) { 
    
    #ifdef LOG_LINK_DROP
      printf("[link-drop-recv] at %8.4f < ag %d nd %d > from < ag %d nd %d >\n",Scheduler::Clock(),id,n->id,mm->ainfo.agent_id,mm->ainfo.node_id);
    #endif // LOG_LINK_DROP
    
    purge_from_neighbor_list(mm);
    cleanup_rt(mm->ainfo.agent_id);
  }
  return; 
}  

void mcAgent::specific_send_data_pkt (void) {

  assert (started == true);

  #ifdef LOG_DATA
    printf ("[data] at %8.4f < ag %d nd %d > source-data-pkt < %ld > ssn\n", Scheduler::Clock(), id, n->id, data_seq_no);
  #endif // LOG_DATA

  tx_data();
  data_seq_no++; 
  return;
}

void mcAgent::tx_data() {

  for (void *pos = mesh_neighbor_list.GetHeadPosition();
       pos != NULL;
       mesh_neighbor_list.GetNext(pos) ) {

    AppPacket *ap = new AppPacket(DATA); 
    ap->u.data_p.src_agent_id = id; 
    ap->u.data_p.src_node_id = n->id; 
    ap->u.data_p.seq_no = data_seq_no; 
    MeshMember *mm = mesh_neighbor_list.GetAt(pos);

    #ifdef LOG_DATA 
      printf("[data-send] at %8.4f < ag %d nd %d > sent-data-to < ag %d nd %d > < %ld > ssn \n", Scheduler::Clock(), id, n->id, mm->ainfo.agent_id, mm->ainfo.node_id, data_seq_no); 
    #endif // LOG_DATA 

    send_pkt(ap,mm->ainfo.agent_id,mm->ainfo.node_id); 
  }
}



void mcAgent::handle_data(AppPacket *ap) {

  MeshMember *mm = find_mesh_member_info(ap->src_agent);
  long diff; 
 
  if ((mm != NULL) && (mm->is_neighbor)) {
    MeshMember *src_mm = find_mesh_member_info(ap->u.data_p.src_agent_id); 
    if (src_mm != NULL) {
      diff = ap->u.data_p.seq_no - src_mm->last_data_seq_no; 
      if (diff > 0) {
        src_mm->last_data_seq_no = ap->u.data_p.seq_no; 

        #ifdef LOG_DATA 
	  {
          printf("[data-recv] at %8.4f < ag %d nd %d > from < ag %d nd %d > originating-at < ag %d nd %d > < %ld > ssn\n",Scheduler::Clock(), id, n->id, ap->src_agent, ap->src, src_mm->ainfo.agent_id, src_mm->ainfo.node_id, src_mm->last_data_seq_no);  
	  }
        #endif // LOG_DATA

        #ifdef LOG_DATA_LOSS
          {
          if (diff > 1) 
            printf("[data-loss] at %8.4f < ag %d nd %d > detects < %ld > lost-pkts\n",Scheduler::Clock(), id, n->id, diff-1);
	  }
	#endif // LOG_DATA_LOSS

        for (int i = 0; i < MAXMESHMEMBERS; i++) {
          if (src_mm->source[i] == true) {
            MeshMember *dest = find_mesh_member_info(i); 
            if (dest != NULL) {
              if (dest->is_neighbor == true) {
                AppPacket *app = new AppPacket(DATA); 
                app->u.data_p.src_agent_id = ap->u.data_p.src_agent_id; 
                app->u.data_p.src_node_id = ap->u.data_p.src_node_id; 
        	app->u.data_p.seq_no = ap->u.data_p.seq_no; 
                
                #ifdef LOG_DATA 
		  printf("[data-forward] at %8.4f < ag %d nd %d > is-src < ag %d nd %d > is-dest < ag %d nd %d > is-origin < %ld > is ssn\n", Scheduler::Clock(), id, n->id, dest->ainfo.agent_id, dest->ainfo.node_id, ap->u.data_p.src_agent_id, ap->u.data_p.src_node_id, ap->u.data_p.seq_no); 
    		#endif // LOG_DATA 
              
		send_pkt(app,dest->ainfo.agent_id,dest->ainfo.node_id); 
	      }
            }
          }
        }
      }
    }
  }  
}

void mcAgent::handle_mesh_rt_exchange_timeout() {

 for (void *pos = mesh_neighbor_list.GetHeadPosition();
       pos != NULL;
       mesh_neighbor_list.GetNext(pos) ) {

    MeshMember *mm = mesh_neighbor_list.GetAt(pos);
    if (mm->gotrt == false) {
      //purge_from_neighbor_list(mm);
      cleanup_rt(mm->ainfo.agent_id);

      #ifdef LOG_TIMEOUTS
        printf("[rtx-timeout] at %8.4f < ag %d nd %d > detects-for < ag %d nd %d >\n",Scheduler::Clock(),id,n->id,mm->ainfo.agent_id,mm->ainfo.node_id);
      #endif // LOG_TIMEOUTS

    }
    else mm->gotrt = false; 
  }
  return; 
}

void mcAgent::handle_mesh_setup_neighbor_timeout(MeshMember *mm) {

  void *pos = mesh_setup_neighbor_timer_list.Locate(mm->ainfo.agent_id); 

/* Suman's correction */
  assert (pos != NULL);
  // printf ("[timer-debug] Timeout happened at %8.3f for %d at %d\n", Scheduler::Clock(), mm->ainfo.agent_id, id);

  MeshSetupNeighborTimer *timer = mesh_setup_neighbor_timer_list.GetAt(pos); 
  delete timer;
  mesh_setup_neighbor_timer_list.RemoveAt(pos); 

  #ifdef LOG_TIMEOUTS
    printf("[setup-neigh-timeout] at %8.4f < ag %d nd %d > detects < ag %d nd %d >\n",Scheduler::Clock(),id,n->id,mm->ainfo.agent_id,mm->ainfo.node_id); 
  #endif // LOG_TIMEOUTS

  purge_from_setup_neighbor_list(mm); 
  return; 
}
  
void mcAgent::cleanup_rt(int lostagent) {
  
  if ((meshrt.pathlen[lostagent] > 1) && (meshrt.valid[lostagent])) {
    free(meshrt.path[lostagent]);
    meshrt.path[lostagent] = NULL;
  }
  meshrt.valid[lostagent] = false;
  for (int i = 0; i < MAXMESHMEMBERS; i++) {
    if ((meshrt.valid[i]) && (meshrt.pathlen[i] > 1) && (meshrt.path[i][0].agent_id = lostagent)) {
      meshrt.valid[i] = false;       
      free(meshrt.path[i]);
      meshrt.path[i] = NULL; 
    }
  }
  return; 
}
//

void mcAgent::handle_mesh_member_refresh (AppPacket *ap) {

  // Ignore message originating from self
  if (ap->u.meshrefresh_p.origin.agent_id == id)
    return;

  MeshMember *mm = find_mesh_member_info (ap->src_agent);
  if (mm == NULL) // Do not accept anything from non-neighbors
    return;
  if (mm->last_seq_no >= ap->u.meshrefresh_p.neighbor_seq_no)
    return;

  if (ap->u.meshrefresh_p.is_alive)
    handle_mesh_member_refresh_alive(ap);
  else
    handle_mesh_member_refresh_dead(ap);
  return;
}

void mcAgent::remove_mesh_member (int ag_id) {

  void *pos = all_agent_list.Locate(ag_id);
  assert (pos != NULL);
  MeshMember *mm = all_agent_list.GetAt(pos);
  all_agent_list.RemoveAt(pos);
  if (mm->neighbor_setup_pending == true) {
    void *setup_pos = setup_neighbor_list.Locate(ag_id);
    assert (setup_pos != NULL);
    setup_neighbor_list.RemoveAt(setup_pos);
  }
  if (mm->is_neighbor == true) {
    void *neighbor_pos = mesh_neighbor_list.Locate(ag_id);
    assert (neighbor_pos != NULL);
    mesh_neighbor_list.RemoveAt(neighbor_pos);
  }
  if (mm->is_stopped == true) {
    void *stop_pos = stopped_member_list.Locate(ag_id);
    assert (stop_pos != NULL);
    stopped_member_list.RemoveAt(stop_pos);
  }
 return;
}

// Handle information about a dead member
void mcAgent::handle_mesh_member_refresh_dead (AppPacket *ap) {

  MeshMember *mm = find_mesh_member_info(ap->u.meshrefresh_p.origin.agent_id);
  if (mm == NULL)
    return;

  // If this is a valid refresh message ....
  if (ap->u.meshrefresh_p.origin_seq_no <= mm->last_seq_no)
    return;

  if (mm->is_neighbor == true) {

    purge_from_neighbor_list(mm);
    //sri's code
    cleanup_rt(mm->ainfo.agent_id);
    //
    add_to_setup_neighbor_list(mm);
    MeshSetupNeighborTimer *msnt = new MeshSetupNeighborTimer(this,mm);
    msnt->SetTimer(CONST_MESH_SETUP_NEIGHBOR_TIMEOUT);

/* Suman's correction */
      // printf ("[timer-debug] About to add timer at %8.3f for %d at %d\n", Scheduler::Clock(), mm->ainfo.agent_id, id);

    mesh_setup_neighbor_timer_list.Add(msnt, mm->ainfo.agent_id); 
    tx_mesh_neighbor_setup_req(mm);
  }
  if (mm->is_stopped == false)
    add_to_stopped_member_list(mm);

  mm->last_seq_no = ap->u.meshrefresh_p.origin_seq_no;

  #ifdef LOG_APP_REFRESH
  printf ("[mesh-member] at %8.4f < ag %d nd %d > detects-lost-member < ag %d nd %d >\n", Scheduler::Clock(), id, n->id, mm->ainfo.agent_id, mm->ainfo.node_id);
  #endif // LOG_APP_REFRESH

  forward_mesh_refresh(mm->ainfo.agent_id,mm->ainfo.node_id, ap->u.meshrefresh_p.origin_seq_no,ap->src_agent,false);
  return;
}

void mcAgent::handle_mesh_member_refresh_alive (AppPacket *ap) {

  MeshMember *member = find_mesh_member_info (ap->u.meshrefresh_p.origin.agent_id);
  if (member == NULL) {
    member = create_new_mesh_member_info(ap->u.meshrefresh_p.origin.agent_id,ap->u.meshrefresh_p.origin.node_id);
  }

  // If this is a valid refresh message ....
  if (ap->u.meshrefresh_p.origin_seq_no > member->last_seq_no) {

    if (member->is_stopped == true)
      purge_from_stopped_member_list(member);
    member->last_seq_no = ap->u.meshrefresh_p.origin_seq_no;
    member->last_refresh = Scheduler::Clock(); 
    forward_mesh_refresh(member->ainfo.agent_id,member->ainfo.node_id, ap->u.meshrefresh_p.origin_seq_no,ap->src_agent,true);
  }
return;
}

MeshMember *mcAgent::find_mesh_member_info (int aid) {

  void *pos = all_agent_list.Locate(aid);
  if (pos == NULL)
    return NULL;
  else
    return all_agent_list.GetAt(pos);
}

void mcAgent::add_to_setup_neighbor_list (MeshMember *mm) {

  assert (mm->neighbor_setup_pending == false);
  assert (setup_neighbor_list.Locate(mm->ainfo.agent_id) == NULL);
  mm->neighbor_setup_pending = true;
  setup_neighbor_list.Add(mm,mm->ainfo.agent_id);
  return;
}

void mcAgent::purge_from_setup_neighbor_list (MeshMember *mm) {

  assert (mm->neighbor_setup_pending == true);
  mm->neighbor_setup_pending = false;
  void *pos = setup_neighbor_list.Locate(mm->ainfo.agent_id);
  assert (pos != NULL);
  setup_neighbor_list.RemoveAt(pos);
  return;
}

void mcAgent::add_to_stopped_member_list (MeshMember *mm) {

  assert (mm->is_stopped == false);
  assert (stopped_member_list.Locate(mm->ainfo.agent_id) == NULL);
  mm->is_stopped = true;
  stopped_member_list.Add(mm,mm->ainfo.agent_id);
  return;
}

void mcAgent::purge_from_stopped_member_list (MeshMember *mm) {

  assert (mm->is_stopped == true);
  mm->is_stopped = false;
  void *pos = stopped_member_list.Locate(mm->ainfo.agent_id);
  assert (pos != NULL);
  stopped_member_list.RemoveAt(pos);
  return;
}

void mcAgent::add_to_neighbor_list (MeshMember *mm) {

  #ifdef LOG_APP_NEIGHBOR
    printf ("[mesh-neighbor] at %8.4f < ag %d nd %d > adds-new-neighbor < ag %d nd %d >\n", Scheduler::Clock(), id, n->id, mm->ainfo.agent_id, mm->ainfo.node_id);
  #endif // LOG_APP_NEIGHBOR

  assert (mm->is_neighbor == false);
  assert (mesh_neighbor_list.Locate(mm->ainfo.agent_id) == NULL);
  mm->is_neighbor = true;
  mesh_neighbor_list.Add(mm,mm->ainfo.agent_id);
  //sri's code
  meshrt.cost[mm->ainfo.agent_id] = n->rt->get_path_cost(mm->ainfo.node_id);
  meshrt.valid[mm->ainfo.agent_id] = true; 
  meshrt.pathlen[mm->ainfo.agent_id] = 1; 
  meshrt.path[mm->ainfo.agent_id] = NULL;
  mm->gotrt = true;
  mm->consensuscost = LINKDROPTHRESHOLD + 1;   
  //
  return;
}

void mcAgent::purge_from_neighbor_list (MeshMember *mm) {

  assert (mm->is_neighbor == true);
  mm->is_neighbor = false;
  void *pos = mesh_neighbor_list.Locate(mm->ainfo.agent_id);
  assert (pos != NULL);
  mesh_neighbor_list.RemoveAt(pos);

  #ifdef LOG_PURGE_NEIGHBOR
    printf("[purge-neigh] at %8.4f < ag %d nd %d > removes < ag %d nd %d >\n",Scheduler::Clock(),id,n->id,mm->ainfo.agent_id,mm->ainfo.node_id);
  #endif // LOG_PURGE_NEIGHBOR

  return;
}

// Process the neighbor setup request
void mcAgent::handle_mesh_neighbor_setup_req (AppPacket *ap) {

  bool accept = true;
  MeshMember *mm = find_mesh_member_info(ap->src_agent);
  if (mm == NULL) {
    mm = create_new_mesh_member_info(ap->src_agent,ap->src);
  }
  else {
    mm->last_refresh = Scheduler::Clock();
    if (mm->neighbor_setup_pending == true) {
      void *pos = mesh_setup_neighbor_timer_list.Locate(mm->ainfo.agent_id); 
      assert (pos != NULL);
      MeshSetupNeighborTimer *timer = mesh_setup_neighbor_timer_list.GetAt(pos); 
      timer->CancelTimer(); 
      delete timer;
      mesh_setup_neighbor_timer_list.RemoveAt(pos); 
      purge_from_setup_neighbor_list(mm);
    }
    if (mm->is_stopped == true) {
      purge_from_stopped_member_list(mm);
    }
  }

  if (accept == true) {
    if (mm->is_neighbor == false) {
      add_to_neighbor_list(mm);
    }
  }
  tx_mesh_neighbor_setup_resp(mm,accept);
  return;
}
      
// Process the neighbor setup response
void mcAgent::handle_mesh_neighbor_setup_resp (AppPacket *ap) {     
   
  MeshMember *mm = find_mesh_member_info(ap->src_agent);
  if (mm == NULL) {
    mm = create_new_mesh_member_info(ap->src_agent,ap->src);
  }

  void *pos = mesh_setup_neighbor_timer_list.Locate(ap->src_agent); 

  /* Suman's correction */
  // printf ("[timer-debug] About to delete timer at %8.3f for %d at %d\n", Scheduler::Clock(), ap->src_agent, id);

  if (pos != NULL) {
    MeshSetupNeighborTimer *timer = mesh_setup_neighbor_timer_list.GetAt(pos);  
    timer->CancelTimer(); 
    delete timer; 
    mesh_setup_neighbor_timer_list.RemoveAt(pos); 
  }  

  if (mm->neighbor_setup_pending == true) {
    // Act only if neighbor setup is pending
    purge_from_setup_neighbor_list(mm);
    if (mm->is_stopped == true)
      purge_from_stopped_member_list(mm);
    assert (mm->is_neighbor == false);
    if (ap->u.meshneighborsetupresp_p.accept == true) {

      #ifdef LOG_APP_NEIGHBOR
        printf ("[mesh-neighbor] at %8.4f < ag %d nd %d > detects-new-neighbor < ag %d nd %d >\n", Scheduler::Clock(), id, n->id, mm->ainfo.agent_id, mm->ainfo.node_id);
      #endif // LOG_APP_NEIGHBOR

      add_to_neighbor_list(mm);
      mm->last_refresh = Scheduler::Clock();
    }
  }
  return;
}

void mcAgent::handle_mesh_refresh_timeout (void) {

  for (void *pos = all_agent_list.GetHeadPosition();
       pos != NULL;
       all_agent_list.GetNext(pos) ) {

    MeshMember *mm = all_agent_list.GetAt(pos);
    if (mm->last_refresh < (Scheduler::Clock() - CONST_MESH_NEIGHBOR_REFRESH_TIMEOUT)) {
      // This member is lost ...
      if (mm->is_stopped == false) {
	add_to_stopped_member_list(mm);
        //sri's code
        cleanup_rt(mm->ainfo.agent_id);
      }

      if (mm->is_neighbor == true) {

        #ifdef LOG_APP_REFRESH
	  printf ("[mesh-neighbor] at %8.4f < ag %d nd %d > detects-lost-neighbor < ag %d nd %d >\n", Scheduler::Clock(), id, n->id, mm->ainfo.agent_id, mm->ainfo.node_id);
	#endif // LOG_APP_REFRESH

	purge_from_neighbor_list(mm);
	// This neighbor is lost ... so inform all other members
	// forward_mesh_refresh(mm->ainfo.agent_id,mm->ainfo.node_id, -1 ,-1,false);
	mm->last_seq_no ++;
	forward_mesh_refresh(mm->ainfo.agent_id,mm->ainfo.node_id, mm->last_seq_no,-1,false);
	// If number of neighbors is low, try to add some new ones ?
      }

      if (mm->neighbor_setup_pending == true) {
        void *pos = mesh_setup_neighbor_timer_list.Locate(mm->ainfo.agent_id); 
        assert (pos != NULL);
        MeshSetupNeighborTimer *timer = mesh_setup_neighbor_timer_list.GetAt(pos); 
        timer->CancelTimer(); 
        delete timer;
        mesh_setup_neighbor_timer_list.RemoveAt(pos); 
	purge_from_setup_neighbor_list(mm);
      }
    }
  }

  /* Reset the timer */
  mesh_refresh_timer->SetTimer(CONST_MESH_NEIGHBOR_REFRESH_TIMEOUT);
  return;
}

void mcAgent::handle_stopped_member_timeout (void) {

}

//sri's code

void MeshRefreshTimeoutTimer::EventHandler (void) {

  mca->handle_mesh_refresh_timeout(); 
  return;
}

void MeshKeepAliveTimer::EventHandler (void) {

  mca->tx_mesh_refresh ();
  mca->mesh_keepalive_timer->SetTimer(CONST_MESH_NEIGHBOR_KEEPALIVE_PERIOD);
  return;
}

//sri's code 
void MeshLinkAddTimer::EventHandler (void) {

  mca->tx_mesh_link_add ();
  mca->mesh_link_add_timer->SetTimer(CONST_MESH_LINK_ADD_PERIOD);
  return;
}

void MeshLinkDropTimer::EventHandler (void) {

  mca->tx_mesh_link_drop ();
  mca->mesh_link_drop_timer->SetTimer(CONST_MESH_LINK_DROP_PERIOD);
  return;
}

void MeshRTExchangeTimer::EventHandler (void) {

  mca->tx_mesh_rt_exchange ();
  mca->mesh_rt_exchange_timer->SetTimer(CONST_MESH_RT_EXCHANGE_PERIOD);
  return;
}

void MeshRTExchangeCheckTimer::EventHandler (void) {

  mca->handle_mesh_rt_exchange_timeout ();
  mca->mesh_rt_exchange_check_timer->SetTimer(CONST_MESH_RT_EXCHANGE_TIMEOUT);
  return; 
} 

void MeshSetupNeighborTimer::EventHandler (void) {

  mca->handle_mesh_setup_neighbor_timeout(mm); 
  return; 
}

void DataTimer::EventHandler (void) {
  
  mca->tx_data(); 
  mca->data_timer->SetTimer(mca->get_time_till_next_transmit()); 
  return; 
}
//

void StoppedMemberTimer::EventHandler (void) {

  mca->handle_stopped_member_timeout();
  return;
}

double mcAgent::get_time_till_next_transmit (void) {

  return CONST_DATA_TRANSMIT_INTERVAL; 
}



















