/**************************************************
 *
  File: common-agent.cc
  Author: Suman Banerjee <suman@cs.umd.edu>
  Date: July 31, 2001
  Terms: GPL

  NICE implementation in myns

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 **************************************************/

#include <stdio.h>
#include <assert.h>
#include "nicenode.h"
#include "common-agent.h"
#include <scheduler.h>

#undef LOG_DETAILED_PKT_SEND
#undef LOG_DETAILED_PKT_RCV

void commonAgent::send_pkt_wrapper (AppPacket *ap, int aid, int nid) {

#ifdef LOG_DETAILED_PKT_SEND
   printf ("[ %d %d ] spd %f [ %d %d ] tp %d", id, n->id, Scheduler::Clock(), aid, nid, ap->st);
  display_detailed_pkt_info(ap);
#endif // LOG_DETAILED_PKT_SEND
  send_pkt(ap,aid,nid);
}

void commonAgent::rx_pkt_wrapper (AppPacket *ap) {
#ifdef LOG_DETAILED_PKT_RCV
   printf ("[ %d %d ] rpd %f [ %d %d ] tp %d", id, n->id, Scheduler::Clock(), ap->src_agent, ap->src, ap->st);
  display_detailed_pkt_info(ap);
#endif // LOG_DETAILED_PKT_RCV
  return;
}

void commonAgent::display_detailed_pkt_info (AppPacket *ap) {

  switch (ap->st) {

  case JOIN_QUERY:
    printf (" sz 5 + 0\n");
    break;

  case JOIN_FORWARD:
    printf (" sz %d + 0\n", 2 * HOST_INFO_SIZE +  9);
    break;

  case PING_QUERY:
    printf (" sz 5 + 0\n");
    break;

  case JOIN_RESPONSE:
    printf (" sz %d + %d\n", HOST_INFO_SIZE + 7, HOST_INFO_SIZE * ap->u.joinresp_p.mbr_count);
    break;

  case CLUSTER_REFRESH:
    printf (" sz 13 + %d\n", HOST_INFO_SIZE * (ap->u.clusterrefresh_p.mbr_count+ap->u.clusterrefresh_p.hl_mbr_count));
    break;

  case CLUSTER_MERGE:
    printf (" sz 5 + %d\n", HOST_INFO_SIZE * ap->u.clustermerge_p.mbr_count);
    break;

  case PING_RESPONSE:
    printf (" sz 9 + %d\n", HOST_INFO_SIZE * ap->u.pingresp_p.mbr_count);
    break;

  case PACKET_DATA:
    printf(" sz %d + 0\n", 2 + HOST_INFO_SIZE);
    break;

  default:
    
    assert(0);
  }

  return;
}
